
;===================================================================
;
; PROGRAM:  mollosvd_make
; 
; PURPOSE:  generates "super-channel" maps from simulation data cubes.
;           It's not restricted to molecular lines. Source functions
;           and absorption coefficient can be modified.
; INPUT  :  Data cubes (density, velocity, temperature). Add
;             masks for tracer location as needed.
;           Leiden Molecular Database for transitions etc. 
; OUTPUT :  Fits file (vlos,xpos,ypos)
;           Fits files are channel map files and can be directly
;           viewed with ds9. 
; USAGE  :  This is a multi-step process
;           (1) use mollosvd_make to generate "super-channel" maps,
;           i.e.high-resolution ppv cubes. These are the raw data and
;           are generally at higher spectral (and spatial) resolution
;           than final products.
;           filename: mollosvd{axis}_{emitter}_{dump#}.fits
;           filename: mollosvd{axis}_{emitter}_{dump#}.head
;           (2) these ppv-cubes can be degraded to "observations",
;           using mollosvd_trans[late]. Degrading in velocity space
;           is done by binning, degrading in position space uses 
;           a Gaussian filter.
;           filename: mollosvd{axis}_{emitter}_{dump#}_c{channel#}s{beamwidth}e{noise}.fits
;           The beamwidth is in pixels, and the noise level is in K.
;           Can be used to read rotated dumps of rotatehvc. Box sizes
;           need to be adapted then. Rotated dumps only read if array
;           angle is defined.
;           n0 and T0 are scaling variables necessary for correct unit conversion.
; NOTES   : ilo:ihi and jlo:jhi area ALWAYS in the plane of sky.
;           Most of the RT part is from Rybicki & Lightman, and Shu.
;           Assumes Rayleigh-Jeans limit, i.e converts to brightness temperature
;           Uses Leiden Molecular Database for transitions etc. Number
;           of levels included in computation can be chosen. For
;           testing purposes, 5 levels are practical on a 256x256 grid.
;           Currently, only optically thin lines in non-LTE, i.e. 
;           self-absorption is not handled correctly, since we don't 
;           follow the emitted photons to determine the new population
;           levels (would involve iterating). In other words, assumes
;           LVG approximation (each channel is optically thin). 
;--------------------------------------------------------------------

pro mollosvd_make

  descr = 'c3f0g3b'

  if (descr eq 'c3f0g3b') then begin; core right of center
    ndim   = 3
    legend0= '!6c3f0g3b'
    models = 2
    mdl    = 'c3f0g3b_1'
    nres   = [200l,404l,404l]
    dmp    = [25,25]
    idir   = 2
    ilo    = lonarr(models)
    ihi    = lonarr(models)
    jlo    = lonarr(models)
    jhi    = lonarr(models)
    klo    = long(0)
    khi    = long(nres[idir-1]-1)
    angle  = [60,75]
    leng    = [2.1075e-2,4.4345e-2,4.4345e-2]*2.0 
    ciso    = 0.0 ; in Kelvin
    n0      = 1e-2
    T0      = 1e4
    whichmol= [6]
    ishield = [1,1,1,1,1,1,0] ; if 1, use irrad dumps and shld field in molecule definition
    if (idir eq 2) then begin
      icen   = long(90)
      jcen   = long(150)
      iwidth = 60l
      jwidth = 100l
    endif
    nchannel = 120l
    minvph = -16.0 ; in km/s
    maxvph = 32.0
    dtdmp    = 2.16e-3
    disk     = '~/proteus/run/hvc/'
  endif

  if (descr eq 'Gf2') then begin; core right of center
    ndim   = 3
    legend0= '!6Gf2'
    models = 1
    mdl    = '3aaww00d0d1_g'
    nres   = [256l,512l,512l]
    dmp    = [19]
    idir   = 3
    ilo    = lonarr(models)
    ihi    = lonarr(models)
    jlo    = lonarr(models)
    jhi    = lonarr(models)
    klo    = long(0)
    khi    = long(nres[idir-1]-1)
;    angle  = 1l
    leng    = [2.5,5.0,5.0]
    ciso    = 0.0 ; in Kelvin
    n0      = 1.0
    T0      = 1.0
    whichmol= [0,1,2]
    ishield = [1,1,1,1,1,1] ; if 1, use irrad dumps and shld field in molecule definition
    if (idir eq 1) then begin
      icen   = long(256)
      jcen   = long(256)
      iwidth = 63l
      jwidth = 63l
    endif
    if (idir eq 2) then begin
      icen   = long(128)
      jcen   = long(256)
      iwidth = 63l
      jwidth = 63l
    endif
    if (idir eq 3) then begin
      icen   = long(128)
      jcen   = long(256)
      iwidth = 127l
      jwidth = 127l
    endif
    nchannel = 40l
    minvph = -2.5 ; in km/s
    maxvph = 2.5
    dtdmp    = 8e-3
    disk     = '~/proteus/run/collimflow/'
  endif

  ctable   = 39

  convvelo = 9.118e-2*sqrt(T0)      ;[km s^(-1)]
  convtime = 9.528e1 /sqrt(n0)      ;[Myr]
  convleng = 2.74e19 *sqrt(T0/n0)   ;[cm]
  convpars = 8.896e0 *sqrt(T0/n0)   ;[pc]
  convmass = 1.72e1  *sqrt(T0^3/n0) ;[Msol]
  convsurf = 2.74e19 *sqrt(T0*n0)
  convkmsc = 1e5 ; no scalings, because it's applied to already scaled values
  convintb = 9d20/(2d0*1.381d-16) ; for brightness Temperature

; constants
  pi     = 4d0*atan(1d0)
  clight = 3.0d10
  kboltz = 1.381d-16
  hplanck= 6.626d-27

;---------------------------------------------------------------------
; the molecule "database"
; molecule transition identifies (MOL-TRANS) must be separated by a
; dash for a correct interpretation by molbase_read.pro
  nallmol     = 6                    ; length of molecular database. This is one less than entries (HI)
  nmol        = n_elements(whichmol) ; total number of used molecules
  nmollev     = 5                    ; number of levels for LTE populations (max=10)
; Fields in molecule structure:
; constants:
;   fillab : file label for output file. Must contain molecule identifier in first part and transition 
;            identifier in 2nd, separated by a "-".
;   mwgt   : molecular weight
;   su     : upper level
;   sl     : lower level
;   nlev   : number of levels to be used for level populations (currently limited to 10)
;   abnd   : abundance (in lieu of chemistry). Combine with mask if needed.
;   shld   : shielding factor as in shld = exp(-tau). Assumes radiation field derived from irrad.x
;            If 0.0, irrad mask is not read.
;   frez   : > 0.0: freeze-out density (i.e. higher densities are not considered)
;   frez   : < 0.0: threshold density
;   frez   : = 0.0: not used
; variable fields
;   ncr    : critical density (will be filled in for each cell depending on population levels). Currently not used.
;   pcoll  : collision rate matrix. Needs to be generated for each
;            velocity channel and cell. See molbase_read.pro and molbase_abnd.pro
;   peinst : Einstein A coefficients, starting at peinst[1]
;   pfreq  : transition frequencies , starting at pfreq[1]
;   pener  : level energy (not difference!)
; data from Tielens, Physics and Chemistry of the ISM; Leiden Molecular Database
; 12CO, (1-0). Abundance from Burgh, France & McCandliss 2007, ApJ 658, 446
  struct_12c16o_10 = {MOL,fillab:'12C16O-10',mwgt:28.0,su:1,sl:0,nlev:nmollev,abnd:3e-6,shld:0.670,frez:1e4,temp:0.0,$
                          ncr:ptr_new(fltarr(1)),pcoll:ptr_new(fltarr(nmollev,nmollev)),$
                          peinst:ptr_new(fltarr(nmollev)),pfreq:ptr_new(fltarr(nmollev)),$
                          pener:ptr_new(fltarr(nmollev))}
  struct_12c16o_21 = {MOL,fillab:'12C16O-21',mwgt:28.0,su:2,sl:1,nlev:nmollev,abnd:3e-6,shld:0.670,frez:1e4,temp:0.0,$
                          ncr:ptr_new(fltarr(1)),pcoll:ptr_new(fltarr(nmollev,nmollev)),$
                          peinst:ptr_new(fltarr(nmollev)),pfreq:ptr_new(fltarr(nmollev)),$
                          pener:ptr_new(fltarr(nmollev))}
  struct_12c16o_32 = {MOL,fillab:'12C16O-32',mwgt:28.0,su:3,sl:2,nlev:nmollev,abnd:3e-6,shld:0.670,frez:1e4,temp:0.0,$
                          ncr:ptr_new(fltarr(1)),pcoll:ptr_new(fltarr(nmollev,nmollev)),$
                          peinst:ptr_new(fltarr(nmollev)),pfreq:ptr_new(fltarr(nmollev)),$
                          pener:ptr_new(fltarr(nmollev))}
; 13CO  (1-0). Abundance from Burgh, France & McCandliss 2007, ApJ 658, 446
  struct_13c16o_10 = {MOL,fillab:'13C16O-10',mwgt:29.0,su:1,sl:0,nlev:nmollev,abnd:5e-8,shld:0.670,frez:1e4,temp:0.0,$
                          ncr:ptr_new(fltarr(1)),pcoll:ptr_new(fltarr(nmollev,nmollev)),$
                          peinst:ptr_new(fltarr(nmollev)),pfreq:ptr_new(fltarr(nmollev)),$
                          pener:ptr_new(fltarr(nmollev))}
  struct_13c16o_21 = {MOL,fillab:'13C16O-21',mwgt:29.0,su:2,sl:1,nlev:nmollev,abnd:5e-8,shld:0.670,frez:1e4,temp:0.0,$
                          ncr:ptr_new(fltarr(1)),pcoll:ptr_new(fltarr(nmollev,nmollev)),$
                          peinst:ptr_new(fltarr(nmollev)),pfreq:ptr_new(fltarr(nmollev)),$
                          pener:ptr_new(fltarr(nmollev))}
  struct_13c16o_32 = {MOL,fillab:'13C16O-32',mwgt:29.0,su:3,sl:2,nlev:nmollev,abnd:5e-8,shld:0.670,frez:1e4,temp:0.0,$
                          ncr:ptr_new(fltarr(1)),pcoll:ptr_new(fltarr(nmollev,nmollev)),$
                          peinst:ptr_new(fltarr(nmollev)),pfreq:ptr_new(fltarr(nmollev)),$
                          pener:ptr_new(fltarr(nmollev))}
; N2H+, (1-0), Abundance from Bergin, private comm.
;  struct_n2hp_10   = {MOL,fillab:'N2Hp-10',mwgt:29.0,su:1,sl:0,nlev:nmollev,abnd:1e-5,shld:0.0,frez:-1e4,$
;                          ncr:ptr_new(fltarr(1)),pcoll:ptr_new(fltarr(nmollev,nmollev)),$
;                          peinst:ptr_new(fltarr(nmollev)),pfreq:ptr_new(fltarr(nmollev)),$
;                          pener:ptr_new(fltarr(nmollev))}
;  struct_n2hp_32   = {MOL,fillab:'N2Hp-32',mwgt:29.0,su:3,sl:2,nlev:nmollev,abnd:1e-5,shld:0.0,frez:-1e4,$
;                          ncr:ptr_new(fltarr(1)),pcoll:ptr_new(fltarr(nmollev,nmollev)),$
;                          peinst:ptr_new(fltarr(nmollev)),pfreq:ptr_new(fltarr(nmollev)),$
;                          pener:ptr_new(fltarr(nmollev))}
; HI must be last in database and needs special treatment
  struct_h21cm     = {MOL,fillab:'HI-21cm',mwgt:1.0,su:1,sl:0,nlev:2,abnd:1.0,shld:0.0,frez:0.0,temp:1e4,$
                          ncr:ptr_new(fltarr(1)),pcoll:ptr_new(fltarr(nmollev,nmollev)),$
                          peinst:ptr_new(fltarr(nmollev)),pfreq:ptr_new(fltarr(nmollev)),$
                          pener:ptr_new(fltarr(nmollev))}
  (*(struct_h21cm.pfreq))[0] = 0.0
  (*(struct_h21cm.pfreq))[1] = 1.420406e9 ; in Hz (molbase_read does conversion for read-in data)

; other molecules to do: C18O (abundance from Frerkin et al. 1982)?,
; NH3, HCO+, CS, H2CO, HCN. For those we had abnd = 1e-8

  molcat    = replicate(struct_12c16o_10,nallmol+1)
  molcat[1] = struct_12c16o_21
  molcat[2] = struct_12c16o_32
  molcat[3] = struct_13c16o_10
  molcat[4] = struct_13c16o_21
  molcat[5] = struct_13c16o_32
;  molcat[6] = struct_n2hp_10
;  molcat[7] = struct_n2hp_32
  molcat[6] = struct_h21cm

  !PATH = !PATH + ":~/lib/astrolib"

; read the Leiden Molecular Database
; The format is slighly changed, and only first 10 levels are included.
  for imol = 0,nallmol-1 do begin  
    molbase_read,molcat[imol]
    print,"MOLLOSVD_MAKE: molecule ",molcat[imol].fillab," read. su = ",molcat[imol].su
  endfor

;  for n=0,molcat[0].nlev-1 do begin
;    print,'A_ul = ',(*(molcat[0].peinst))[n],' nu = ',(*(molcat[0].pfreq))[n], ' Eu = ',(*(molcat[0].pener))[n]
;  endfor
;  print,'upper collision matrix: '
;  print,(*(molcat[0].pcoll))
; assume T=10K
;  tass   = 10.0
;  dass   = 1e3
;  abund  = fltarr(nmollev)
;  molbase_abnd,dass,tass,molcat[0],abund
;  ncr = (*(molcat[0].peinst))[1]/(total((*(molcat[0].pcoll))[1,*]))
;  print,'complete collision matrix: '
;  print,(*(molcat[0].pcoll))
;  print,'critical density : ',(*(molcat[0].ncr))[0]
;  print,total(abund)
;  print,abund
;  stop

;=====================================================================

  isangle = 0
  if (n_elements(angle) gt 0) then isangle = 1
  
  ilo[*] = icen - iwidth
  ihi[*] = icen + iwidth
  jlo[*] = jcen - jwidth
  jhi[*] = jcen + jwidth
  nlos   = khi-klo+1

  minh    = fltarr(nmol)
  maxh    = fltarr(nmol)

  mv      = 0.0
  ntot    = 0.0
  csound  = 0.0
  dvrms   = 0.0
  minnu   = 0.0 ; needs to be computed for each molecule...
  maxnu   = 0.0 
  dnu     = 0.0
  ddnu    = 0.0
  alphnu  = 0.0
  nulos   = 0.0
  nuchan  = 0.0
  tnuchan = 0.0
  nu0     = 0.0
  tmol    = molcat[0]

; for shielding files (irrad dumps)
  timec   = 0.0
  tnx     = 0l
  nindx   = 0l
  comcoord= lonarr(3)
  ix      = 0l
  iy      = 0l
  iz      = 0l
  knumtau = 0l

  dens          = ptr_new(fltarr(nres[0],nres[1],nres[2]))
  vlos          = ptr_new(fltarr(nres[0],nres[1],nres[2]))
  temp          = ptr_new(fltarr(nres[0],nres[1],nres[2]))
  mask          = ptr_new(fltarr(nres[0],nres[1],nres[2]))
  ttdens        = ptr_new(fltarr(nlos))
  ttvlos        = ptr_new(fltarr(nlos))
  tttemp        = ptr_new(fltarr(nlos))       
  ttxpos        = ptr_new(fltarr(nlos))
  tintensnu     = ptr_new(fltarr(nchannel))

  dx = leng[idir-1]*convleng/float(nres[idir-1])
  for k=0,nlos-1 do begin
    (*ttxpos)[k] = (0.5+float(k))*dx
  endfor

  for m=0,models-1 do begin
    ; work on subdomains only, and store line-of-sight direction in fast
    ; (first) index.
    tvlos       = ptr_new(fltarr(nlos,ihi[m]-ilo[m]+1,jhi[m]-jlo[m]+1))
    tdens       = ptr_new(fltarr(nlos,ihi[m]-ilo[m]+1,jhi[m]-jlo[m]+1))
    ttemp       = ptr_new(fltarr(nlos,ihi[m]-ilo[m]+1,jhi[m]-jlo[m]+1))
    intensmapnu = ptr_new(fltarr(ihi[m]-ilo[m]+1,jhi[m]-jlo[m]+1,nchannel,nmol))
    nar = (ihi[m]-ilo[m]+1l)*(jhi[m]-jlo[m]+1l) ; area normalization
    timec = 1.0
;   read in the data sets
    numstr = string(dmp[m],format='(i4.4)')
    if (isangle eq 1) then astr = '_a'+string(angle[m],format='(i3.3)')+'_' else astr = ''
    fnm = disk+mdl+'/dens'+astr+numstr
    print,'MOLLOSVD_MAKE: reading file ',fnm
    proteusread,filename=fnm,dat=(*dens),timec=timec,/single
    case (idir) of
    1:  fnm = disk+mdl+'/momx'+astr+numstr
    2:  fnm = disk+mdl+'/momy'+astr+numstr
    3:  fnm = disk+mdl+'/momz'+astr+numstr
    endcase 
    print,'MOLLOSVD_MAKE: reading file ',fnm
    proteusread,filename=fnm,dat=(*vlos),/single
    fnm = disk+mdl+'/prss'+astr+numstr
    if (ciso eq 0.0) then begin
      print,'MOLLOSVD_MAKE: reading file ',fnm
      proteusread,filename=fnm,dat=(*temp),/single
    endif else begin
      (*temp)[*,*,*] = ciso
    endelse 
    ; for rotated data sets, density etc can be zero. So we have to engineer
    ; the datasets somewhat.
    if (isangle eq 1) then begin 
      knum1 = 0l
      knum2 = 0l
      indx1 = ptr_new(where((*dens) le 1e-15,knum1))
      indx2 = ptr_new(where((*dens) gt 1e-15,knum2))
      if ((knum1 gt 0) and (knum2 gt 0)) then begin
        (*dens)[*indx1] = min((*dens)[*indx2])
        (*temp)[*indx1] = max((*temp)[*indx2])
      endif
    endif
    (*temp) = (*temp)/(*dens)
    (*vlos) = convvelo*(*vlos)/(*dens)
    (*dens) = (*dens)*n0
    (*temp) = (*temp)*T0
    print,format='("MOLLOSVD_MAKE: min/max(dens) = ",e13.5,e13.5)',min(*dens),max(*dens)
    print,format='("MOLLOSVD_MAKE: min/max(temp) = ",e13.5,e13.5)',min(*temp),max(*temp)
    print,format='("MOLLOSVD_MAKE: min/max(vlos) = ",e13.5,e13.5)',min(*vlos),max(*vlos)
    mvlos = total((*vlos)*(*dens))/total((*dens))
    print,format='("MOLLOSVD_MAKE: com vlos      = ",e13.5,e13.5)',mvlos

    help,*dens
    surf  = total((*dens)[ilo[m]:ihi[m],klo:khi,jlo[m]:jhi[m]],2)*dx
    tvscl,surf
    print,min(surf),max(surf)

;    stop

;   loop over the molecules. Since they have different frequencies, we
;   can't reuse the velocity grids even for different transition of
;   one molecule.
    for im=0,nmol-1 do begin
      tmol  = molcat[whichmol[im]]
      nu0   = (*(tmol.pfreq))[tmol.su]
      print,"MOLLOSVD_MAKE: molecule = ",tmol.fillab," at nu0 = ",nu0 
;     default is to use all cells
      (*mask)[*,*,*] = 1.0
;     if there's shielding data, use this to exclude cells.
;     Exlude cells by setting them to small value (0.0 is not a good idea...)
      if (ishield[whichmol[im]] eq 1) then begin
        (*mask)[*,*,*] = 0.0
        fnm = disk+mdl+'/irrad'+numstr+'_01.dmp'
        print,'MOLLOSVD_MAKE: reading file ',fnm
        openr,dun,fnm,/f77,/get_lun
        readu,dun,timec
        readu,dun,tnx
        readu,dun,nindx
        readu,dun,comcoord
        indxarr = ptr_new(lonarr(nindx,3))
        tauarr  = ptr_new(fltarr(nindx))
        readu,dun,(*indxarr)
        readu,dun,(*tauarr)
        free_lun,dun
        for idx = 0l,nindx-1 do begin
          ix = (*indxarr)[idx,0l]-1
          iy = (*indxarr)[idx,1l]-1
          iz = (*indxarr)[idx,2l]-1
          (*mask)[ix,iy,iz] = (*tauarr)[idx]
        endfor
        indxtau = ptr_new(where(((*mask) le tmol.shld) and ((*mask) gt 0.0),knumtau))
        print,format='("MOLLOSVD_MAKE: shielded cells: ",i8)',knumtau
        (*mask)[*,*,*] = 1e-15
        if (knumtau gt 0) then begin
          (*mask)[*indxtau] = 1.0
        endif
        ptr_free,indxarr,tauarr,indxtau
      endif
;     check for freeze-out
      if (tmol.frez gt 0.0) then begin
        indxfrz = ptr_new(where((*dens) gt tmol.frez,knumtau))
        print,format='("MOLLOSVD_MAKE: frozen-out cells: ",i8)',knumtau
        if (knumtau gt 0) then begin
          (*mask)[*indxfrz] = 1e-15
        endif
        ptr_free,indxfrz
      endif
;     check for temperature threshold
      if (tmol.temp gt 0.0) then begin
        indxtmp = ptr_new(where((*temp) gt tmol.temp,knumtau))
        print,format='("MOLLOSVD_MAKE: cells T>Tmol: ",i9)',knumtau
        if (knumtau gt 0) then begin
          (*mask)[*indxtmp] = 1e-15
        endif
        ptr_free,indxtmp
      endif
;     needs to be inside molecule loop, bc we might have different
;     criteria for tracers. This is mainly for rearranging cubes such
;     that the first (fastest) index runs along the line of sight.
      case (idir) of 
      1: begin
        for j=jlo[m],jhi[m] do begin
          for i=ilo[m],ihi[m] do begin
            (*tvlos)[*,i-ilo[m],j-jlo[m]] = (*vlos)[*,i,j]
            (*tdens)[*,i-ilo[m],j-jlo[m]] = (*dens)[*,i,j]*(*mask)[*,i,j]
            (*ttemp)[*,i-ilo[m],j-jlo[m]] = (*temp)[*,i,j]
          endfor
        endfor
      end
      2: begin
        for k=klo,khi do begin
          for j=jlo[m],jhi[m] do begin
            for i=ilo[m],ihi[m] do begin
              (*tvlos)[k,i-ilo[m],j-jlo[m]] = (*vlos)[i,k,j]
              (*tdens)[k,i-ilo[m],j-jlo[m]] = (*dens)[i,k,j]*(*mask)[i,k,j]
              (*ttemp)[k,i-ilo[m],j-jlo[m]] = (*temp)[i,k,j]
            endfor
          endfor
        endfor
      end
      3: begin
        for k=klo,khi do begin
          for j=jlo[m],jhi[m] do begin
            for i=ilo[m],ihi[m] do begin
              (*tvlos)[k,i-ilo[m],j-jlo[m]] = (*vlos)[i,j,k]
              (*tdens)[k,i-ilo[m],j-jlo[m]] = (*dens)[i,j,k]*(*mask)[i,j,k]
              (*ttemp)[k,i-ilo[m],j-jlo[m]] = (*temp)[i,j,k]
            endfor
         endfor
       endfor
      end
      endcase
      dnu   = nu0*(maxvph-minvph)*convkmsc/clight ; frequency range
      minnu = nu0-0.5*dnu
      maxnu = nu0+0.5*dnu
      ddnu  = dnu/double(nchannel) ; frequency increment
      for j=0,jhi[m]-jlo[m] do begin
        for i=0,ihi[m]-ilo[m] do begin
          (*ttdens)[*] = reform((*tdens)[klo:khi,i,j])            
          (*tttemp)[*] = reform((*ttemp)[klo:khi,i,j])           
          (*ttvlos)[*] = reform((*tvlos)[klo:khi,i,j])  
;         this is where all the work is: get intensity (brightness
;         temperature) for each 1D array.
          getintensnu,nlos,nchannel,*ttdens,*tttemp,*ttvlos,*ttxpos,tmol,ddnu,minnu,*tintensnu
          (*intensmapnu)[i,j,*,im] = (*tintensnu)[*]
        endfor
      endfor
;     write to fits file.
      fitshead = strarr(8)
      fitshead[0] = string(ihi[m]-ilo[m]+1,               format='("NAXIS1  = ",i19  ," / [pix]")')
      fitshead[1] = string(jhi[m]-jlo[m]+1,               format='("NAXIS2  = ",i19  ," / [pix]")')
      fitshead[2] = string(nchannel,                      format='("NAXIS3  = ",i19  ," / [chn]")')
      fitshead[3] = string(minvph,                        format='("MINVLS  = ",e20.5," / [km/s]")')
      fitshead[4] = string(maxvph,                        format='("MAXVLS  = ",e20.5," / [km/s]")')
      fitshead[5] = string(min((*intensmapnu)[*,*,*,im]), format='("MINTBR  = ",e20.5," / [K]")')
      fitshead[6] = string(max((*intensmapnu)[*,*,*,im]), format='("MAXTBR  = ",e20.5," / [K]")')
      fnm = disk+mdl+'/mollosvd'+astr+string(idir,format='(i1)')+'_'+tmol.fillab+'_'+numstr+'.fits'
      print,'MOLLOSVD_MAKE: writing file ',fnm
      mwrfits,reform((*intensmapnu)[*,*,*,im]),fnm,fitshead,/CREATE
      help,fitshead
      fitshead = 0
    endfor
    ptr_free,intensmapnu,tdens,tvlos,ttemp
  endfor

;===============================================================

  return
end

;===============================================================
; This function generates the thermal line width for each molecule, 
; used to convolve the velocity information from the simulation grid
; with the thermal broadening. Note that for heavy molecules, the 
; linewidth might still be narrower than the velocity difference
; between cells -- this will be addressed by an adaptive velocity grid
; below. The integral over the profile is 1.
function gaussian,nu,temp,mol
  nu0    = (*(mol.pfreq))[mol.su]
  dnu    = abs(nu-nu0)
  dnud   = 4.30d-7*nu0*sqrt(temp/mol.mwgt)
  return,(5.642d-1)*exp(-(dnu/dnud)^2)/dnud 
end

;===============================================================
; This gets the absorption coefficient alpha_nu (in cm^(-1)),
; ONLY FOR OPTICALLY THIN HI 21cm LINE. CANNOT BE USED FOR
; ANYTHING ELSE. CURRENT CONTROL VIA 2-LEVEL.
; See Kwok p 156
function getalphanu2,nu,dens,temp,mol
  anu = 2.625d-14*(dens/temp)*gaussian(nu,temp,mol)
  return,anu
end

;===============================================================
; This gets the absorption coefficient alpha_nu (in cm^(-1)).
; See Rybicki & Lightman, p. 10, and chapter 1.6
; and Tielens eq. 2.36 combined with Rybicki & Lightman eq. 1.80.
; molbase_abnd determines the population levels in LTE.
function getalphanu,nu,dens,temp,mol
  abund  = fltarr(mol.nlev)
  expo   = exp(-(4.798d-11)*(*(mol.pfreq))[mol.su]/temp)
; population levels from LTE
  molbase_abnd,dens,temp,mol,abund
; abundance of lower level
  nl     = abund[mol.sl]
; the absorption coefficient. See Lyman & Rybicki eq. 1.80
  anu    =   (3.581d19/nu^2) * (2.0*mol.su+1.0)/(2.0*mol.sl+1.0) $
           * (*(mol.peinst))[mol.su]*nl*mol.abnd*(1d0-expo)*gaussian(nu,temp,mol)
  return,anu
end

;===============================================================

function getsource,nu,temp
; this is Planck in frequency space
  return,(1.4724d-47)*nu^3/(exp((4.798d-11)*nu/temp)-1d0)
end


;===============================================================
; This is the main work routine. It determines the intensity along a
; line of sight for all velocity channels.

pro getintensnu,nx,nchannel,dens,temp,vlos,xpos,mol,dnu,minnu,intensnu
  mindvmult = 1
  facdvmult = 3
  k         = 0l
  kmult     = 0l
  km        = 0l
  nuchan    = 0d0
  nulos     = 0d0
  tnuchan   = 0d0
  tintens   = 0d0
  gts       = 0d0
  knum      = 0l
  kmult     = 0l
  nuaxis    = ptr_new(dblarr(nchannel))
  dvlos     = ptr_new(fltarr(nx-1))
  mtemp     = ptr_new(fltarr(nx-1))
  csound    = ptr_new(fltarr(nx-1))
  dvmult    = ptr_new(intarr(nx-1))
  dens0     = 0.0
  temp0     = 0.0
  vlos0     = 0.0
  dens1     = 0.0
  temp1     = 0.0
  vlos1     = 0.0
  nu0       = (*(mol.pfreq))[mol.su]
  maxopa    = -1d30
; loop through the spectrum at a given line of sight
  (*dvlos)[*] = abs(vlos[1:nx-1]-vlos[0:nx-2])*1e5
  (*mtemp)[*] = 0.5*(temp[1:nx-1]+temp[0:nx-2])

; first loop along los: determine the point-to-point velocity
; change. If smaller than thermal line width, mark cells for interpolation
  (*csound)[*]= 1.29d4*sqrt((*mtemp/mol.mwgt))
; the index gives the lower cell 
  indx  = ptr_new(where((*dvlos) gt (*csound),knum))
  (*dvmult)[*] = mindvmult
  if (knum gt 0l) then begin
;   check multiplity by walking through dvlos/csound
;   now check each cell for multiplicity
    for km = 0,knum-1 do begin
      (*dvmult)[(*indx)[km]] = facdvmult*ceil((*dvlos)[(*indx)[km]]/(*csound)[(*indx)[km]])+mindvmult
    endfor
  endif
  ptr_free,indx
  nxnew = fix(total((*dvmult)))+1
  densnew = ptr_new(fltarr(nxnew))
  tempnew = ptr_new(fltarr(nxnew))
  vlosnew = ptr_new(fltarr(nxnew))
  dxnew   = ptr_new(fltarr(nxnew))
  alphanu = ptr_new(fltarr(nxnew))
  xposnew = ptr_new(fltarr(nxnew))
  dtaunew = ptr_new(fltarr(nxnew))
  opacity = ptr_new(fltarr(nxnew))
  kmult = 0l
  kmdv  = 0d0
; (1) generate new positions. Positions are at cell centers! 
; (2) interpolate from old onto new grid
; (3) determine new dx

; (1,2) interpolate and create new grid. We just do linear interpolation
  for k=0,nx-2 do begin
    if ((*dvmult)[k] gt 1) then begin
      for km = 0,(*dvmult)[k]-1 do begin
        kmdv = double(km)/double((*dvmult)[k])
        (*xposnew)[kmult] = xpos[k] + (xpos[k+1]-xpos[k])*kmdv
        (*densnew)[kmult] = dens[k] + (dens[k+1]-dens[k])*kmdv
        (*tempnew)[kmult] = temp[k] + (temp[k+1]-temp[k])*kmdv
        (*vlosnew)[kmult] = vlos[k] + (vlos[k+1]-vlos[k])*kmdv
        kmult = kmult + 1
      endfor
    endif else begin
      (*xposnew)[kmult] = xpos[k]
      (*densnew)[kmult] = dens[k]
      (*tempnew)[kmult] = temp[k]
      (*vlosnew)[kmult] = vlos[k]
      kmult = kmult + 1
    endelse
  endfor
; The last position must be the center position of the last cell.
  (*densnew)[kmult] = dens[nx-1]
  (*tempnew)[kmult] = temp[nx-1]
  (*vlosnew)[kmult] = vlos[nx-1]
  (*xposnew)[kmult] = xpos[nx-1]
; (3) Now generate the new dx-array.
; at the zero position, we have no information about xposnew 
  dx = xpos[1]-xpos[0]
  (*dxnew)[0] = 0.5*dx + 0.5*((*xposnew)[1]-(*xposnew)[0])
  for k=1,nxnew-2 do begin
    (*dxnew)[k] = 0.5*((*xposnew)[k+1]-(*xposnew)[k-1])
  endfor
  (*dxnew)[nxnew-1] = 0.5*((*xposnew)[nxnew-1]-(*xposnew)[nxnew-2]) + 0.5*dx

; take each frequency channel, and do "radtrans" for that.
  for is=0,nchannel-1 do begin
    tintens= 0d0
    nuchan = (0.5d0 + double(is))*dnu+minnu ; current frequency channel
    (*nuaxis)[is] = nuchan
;   1st loop along los: determine alphanu and save it
;   There are two cases here. If we have a 2-level system (currently
;   only optically thin HI), use first branch. Otherwise, do matrix
;   inversion.
;   Branches differ only by one line (getalphanu2 vs getalphanu), but
;   we don't want if-statements in inner loops.
    if (mol.nlev eq 2) then begin
      for k=0,nxnew-1 do begin
;       frequency shift due to bulk motion
        nulos         = nu0*(1d0-(3.33333d-6)*(*vlosnew)[k])  
;       shift channel so that it sees the cell at rest    
        tnuchan       = nuchan + (nu0-nulos)                  
;       Absorption coefficient. Save this, bc we'll need it twice. This is in cm^{-1}
        (*alphanu)[k] = getalphanu2(tnuchan,(*densnew)[k],(*tempnew)[k],mol) 
;       and determine the local (cell-wise) optical depth
        (*dtaunew)[k] = (*alphanu)[k]*(*dxnew)[k]
      endfor
    endif else begin
      for k=0,nxnew-1 do begin
;       frequency shift due to bulk motion
        nulos         = nu0*(1d0-(3.33333d-6)*(*vlosnew)[k])  
;       shift channel so that it sees the cell at rest    
        tnuchan       = nuchan + (nu0-nulos)                  
;       Absorption coefficient. Save this, bc we'll need it twice. This is in cm^{-1}
        (*alphanu)[k] = getalphanu(tnuchan,(*densnew)[k],(*tempnew)[k],mol) 
;       and determine the local (cell-wise) optical depth
        (*dtaunew)[k] = (*alphanu)[k]*(*dxnew)[k]
      endfor
    endelse
;   2nd loop along los:
;   determine opacity between observer and position k. This is Shu eq. 3.9
;   The closer to the observer (k=0), the smaller the opacity.
    (*opacity)[0] = (*dtaunew)[0]
    for k=1,nxnew-1 do begin
      (*opacity)[k] = (*opacity)[k-1] + (*dtaunew)[k]
    endfor
    maxopa = max([maxopa,(*opacity)[nxnew-1]]) ; just for checking
;   3rd loop along los: determine intensity at current frequency.
;   Walk backwards, i.e. towards observer.
;   Since we don't do the full radtrans, the exp(-opacity)
;   term does not add physics (well, it adds the wrong physics).
;   So we split the branch here again:
    if ((*opacity)[nxnew-1] lt 1e-3) then begin ; optically thin branch
      for k=nxnew-1,0,-1 do begin
;       source multiplied by absorption coefficient and attenuated by
;       column towards observer (Shu eq. 3.8 & 3.9)
        gts     = getsource(tnuchan,(*tempnew)[k])
        tintens = tintens + gts*(*dtaunew)[k] ; optically thin
      endfor
    endif else begin
      for k=nxnew-1,0,-1 do begin
;       source multiplied by absorption coefficient and attenuated by
;       column towards observer (Shu eq. 3.8 & 3.9)
        gts     = getsource(tnuchan,(*tempnew)[k])
        tintens = tintens + gts*exp(-(*opacity)[k])*(*dtaunew)[k]; optically thick
      endfor
    endelse
;   the next line assumes Rayleigh-Jeans limit: we're actually
;   returning the brightness temperature (or something like that)
    intensnu[is] = tintens*3.25d36/nuchan^2 
  endfor
  print,format='("GETINTENSNU: max(tau) = ",e13.5,", N = ",e13.5)',maxopa,total((*densnew)*(*dxnew))
  ptr_free,alphanu,nuaxis,dvmult,csound,mtemp,dvlos,densnew,tempnew,vlosnew,dxnew,xposnew,opacity,dtaunew
  return
end


;=====================================================
 
