;===================================================================
;
; PROGRAM:  mollosvd_trans
; 
; PURPOSE:  post-processing (smoothing and down-grading) of
;           the "super-channel" maps generated by mollosvd_make
; INPUT  :  raw channel maps
; OUTPUT :  a "parameter file" and a fits file containing the data.
;           Fits files are channel map files and can be directly
;           viewed with ds9. 
; USAGE  :  This is a multi-step process
;           (1) use mollosvd_make to generate "super-channel" maps,
;           i.e.high-resolution ppv cubes. These are the raw data and
;           are generally at higher spectral (and spatial) resolution
;           than observed data sets.
;           filename: mollosvd{axis}_{emitter}_{dump#}.fits
;           filename: mollosvd{axis}_{emitter}_{dump#}.head
;           (2) these ppv-cubes can be degraded to "observations",
;           using mollosvd_trans[late]. Degrading in velocity space
;           is done by binning, degrading in position space uses 
;           a Gaussian filter.
;           filename: mollosvd{axis}_{emitter}_{dump#}_c{channel#}s{beamwidth}n{noise}.fits
;           filename: mollosvd{axis}_{emitter}_{dump#}_c{channel#}s{beamwidth}n{noise}.head
;           The beamwidth is in pixels, and the noise level is in mJy.
; NOTES   : ilo:ihi and jlo:jhi area ALWAYS in the plane of sky.
;           Most of the RT part is from Rybicki & Lightman
;           Assumes Rayleigh-Jeans limit, i.e converts to brightness temperature
;--------------------------------------------------------------------

pro mollosvd_trans,itest=itest

  descr = 'n128m5p1'

  nangle = 1l

  dtdmp    = 1d-3
  convtime = 9.528d1

  if (descr eq 'n128m5p1') then begin; core right of center
    legend0= '!6n128m5p1'
    models = 5
    mdl    = 'n128m5p1a40b80r00'
    nres   = [128l,128l,128l]
    dmp    = [28,30,32,34,36]
    dump   = [1,1,1,1,1,1]
    idir   = 1
    lthick = [3]
    ltype  = [0]
    ptype  = [14,14,14,14,14,14,14]
    ctype  = [0,16,80,150,208,220,253]
    legend = ["!6n050a40, t = "+string(dtdmp*double(dmp)*convtime,format='(f4.1)')+' Myr']
    squeeze= 0.9
    nangle  = 1l
    leng    = 0.5
    wurst   = 0
    whichmol= [0,1,2]
    width   = 32
    if (idir eq 1) then begin
      icen   = long(64);long(61)
      jcen   = long(64);long(70)
    endif
    if (idir eq 2) then begin
      icen   = long(64)
      jcen   = long(64);long(70)
    endif
    if (idir eq 3) then begin
      icen   = long(64)
      jcen   = long(64)
    endif
    minvph = -2.0e0
    maxvph = 2.0e0
  endif

  nbins    = 80l
  nchannel = 160l
  ctable   = 39

  disk     = '~/proteus/run/taurus/'
  odisk    = '~/proteus/run/taurus/plots/'
  specdisk = '~/proteus/run/taurus/tables/'

  convvelo = 9.118e-2 ;[km s^(-1)]
  convtime = 9.528e1  ;[Myr]
  convleng = 2.74e19  ;[cm]
  convpars = 8.896e0  ;[pc]
  convmass = 1.72e1   ;[Msol]
  convsurf = 2.74e19
  convkmsc = 1e5
  convintb = 9d20/(2d0*1.381d-16) ; for brightness Temperature

  dump[*]  = 1

;---------------------------------------------------------------------
; the molecule "database"
  nallmol     = 3
  nmol        = n_elements(whichmol)
  speclabel = ['!6HCO!e+!n','!6CS','!6HCN']
  filelabel = ['HCOp','CS','HCN']
  
  struct_hco = {MOL, mwgt:29.0, nu0:8.92e10, gu:2, gl:1, a_ul:3.0e-5, ncr:1.7e5, abnd:1e-8}
  struct_cs  = {MOL, mwgt:44.0, nu0:9.80e10, gu:4, gl:2, a_ul:1.7e-5, ncr:3.0e5, abnd:1e-8}
  struct_hcn = {MOL, mwgt:27.0, nu0:8.86e10, gu:2, gl:1, a_ul:2.4e-5, ncr:2.6e6, abnd:1e-8}

  molcat    = replicate(struct_hco,nallmol)
  molcat[1] = struct_cs
  molcat[2] = struct_hcn

  !PATH = !PATH + ":~/lib/astrolib"

;=====================================================================
; constants
  pi     = 4d0*atan(1d0)
  clight = 3.0d10
  kboltz = 1.381d-16
  hplanck= 6.626d-27
  
  ilo[*] = icen - width
  ihi[*] = icen + width
  jlo[*] = jcen - width
  jhi[*] = jcen + width

; plot control
  if (nmol lt 3) then chscl = 1.0 else chscl = 2.4
  xchsz    = 1.0*chscl
  ychsz    = 1.0*chscl
  chsz     = 1.0
  chth     = 3.0

  histarr = fltarr(3,nbins,nmol,models)
  thistarr= fltarr(3,nbins)
  minh    = fltarr(nmol)
  maxh    = fltarr(nmol)

  fwhm    = fltarr(nmol,models)

  mv      = 0.0
  ntot    = 0.0
  csound  = 0.0
  dvrms   = 0.0
  minnu   = 0.0 ; needs to be computed for each molecule...
  maxnu   = 0.0 
  dnu     = 0.0
  ddnu    = 0.0
  alphnu  = 0.0
  nulos   = 0.0
  nuchan  = 0.0
  tnuchan = 0.0
  tmol    = molcat[0]

  dens = ptr_new(fltarr(nres[0],nres[1],nres[2]))
  vlos = ptr_new(fltarr(nres[0],nres[1],nres[2]))
  temp = ptr_new(fltarr(nres[0],nres[1],nres[2]))

  for m=0,models-1 do begin
    nar = (ihi[m]-ilo[m]+1l)*(jhi[m]-jlo[m]+1l) ; area normalization
    dx = leng*convleng/float(nres[idir-1])
    timec = 1.0
    numstr = string(dmp[m],format='(i4.4)')
    if (dump[m] eq 1) then begin
      fnm = disk+mdl+'/dens'+numstr
      print,'MOLLOSVD: reading file ',fnm
      proteusread,filename=fnm,dat=(*dens),timec=timec,/single
      case (idir) of
      1:  fnm = disk+mdl+'/momx'+numstr
      2:  fnm = disk+mdl+'/momy'+numstr
      3:  fnm = disk+mdl+'/momz'+numstr
      endcase 
      print,'MOLLOSVD: reading file ',fnm
      proteusread,filename=fnm,dat=(*vlos),/single
      fnm = disk+mdl+'/prss'+numstr
      if (ciso eq 0.0) then begin
        print,'MOLLOSVD: reading file ',fnm
        proteusread,filename=fnm,dat=(*temp)
        (*temp) = (*temp)/(*dens)
      endif else begin
        (*temp)[*,*,*] = ciso
      endelse 
      (*vlos) = convvelo*(*vlos)/(*dens)
      case (idir) of 
      1: begin
        tvlos = ptr_new((*vlos)[*,ilo[m]:ihi[m],jlo[m]:jhi[m]])
        tdens = ptr_new((*dens)[*,ilo[m]:ihi[m],jlo[m]:jhi[m]])
        ttemp = ptr_new((*temp)[*,ilo[m]:ihi[m],jlo[m]:jhi[m]])
        if keyword_set(itest) then begin
          (*tvlos)[*,*,*] = 0.0
          (*tdens)[*,*,*] = 1e4
          (*tdens)[klo+(khi-klo)/4:khi-(khi-klo)/4-1,*,*] = 1e4
        endif
      end
      2: begin
        tvlos = ptr_new((*vlos)[ilo[m]:ihi[m],*,jlo[m]:jhi[m]])
        tdens = ptr_new((*dens)[ilo[m]:ihi[m],*,jlo[m]:jhi[m]])
        ttemp = ptr_new((*temp)[ilo[m]:ihi[m],*,jlo[m]:jhi[m]])
      end
      3: begin
        tvlos = ptr_new((*vlos)[ilo[m]:ihi[m],jlo[m]:jhi[m],*])
        tdens = ptr_new((*dens)[ilo[m]:ihi[m],jlo[m]:jhi[m],*])
        ttemp = ptr_new((*temp)[ilo[m]:ihi[m],jlo[m]:jhi[m],*])
      end
      endcase
    endif

    case (idir) of 
    1: begin  
      ttdens      = ptr_new(fltarr(nres[idir-1]))
      ttvlos      = ptr_new(fltarr(nres[idir-1]))
      tttemp      = ptr_new(fltarr(nres[idir-1]))       
      tintensnu   = ptr_new(fltarr(nchannel))
      intensmapnu = ptr_new(fltarr(ihi[m]-ilo[m]+1,jhi[m]-jlo[m]+1,nchannel,nmol))
      for im = 0,nmol-1 do begin
        if (dump[m] eq 1) then begin
          tmol  = molcat[whichmol[im]]
          dnu   = tmol.nu0*(maxvph-minvph)*convkmsc/clight
          print,im,dnu
          minnu = tmol.nu0-0.5*dnu
          maxnu = tmol.nu0+0.5*dnu
          ddnu  = dnu/double(nchannel)
          for j=0,jhi[m]-jlo[m] do begin
            for i=0,ihi[m]-ilo[m] do begin
              (*ttdens)[*] = reform((*tdens)[*,i,j])            
              (*tttemp)[*] = reform((*ttemp)[*,i,j])           
              (*ttvlos)[*] = reform((*tvlos)[*,i,j])           
              getintensnu,nres[idir-1],nchannel,*ttdens,*tttemp,*ttvlos,tmol,klo,khi,dx,ddnu,minnu,*tintensnu
              (*intensmapnu)[i,j,*,im] = (*tintensnu)[*]
;              plot,alog10((*ttdens)),(*ttvlos),psym=4
            endfor
          endfor
          fnm = disk+mdl+'/mollosvd'+string(idir,format='(i1)')+'_'+filelabel[im]+'_'+numstr+'.fits'
          print,'MOLLOSVD: writing file ',fnm
          mwrfits,reform((*intensmapnu)[*,*,*,im]),fnm
          fnm = disk+mdl+'/mollosvd'+string(idir,format='(i1)')+'_'+filelabel[im]+'_'+numstr+'.head'
          openw,dun,fnm,/get_lun
          printf,dun,format='("NAXIS1  ",i5)',ihi[m]-ilo[m]+1
          printf,dun,format='("NAXIS2  ",i5)',jhi[m]-jlo[m]+1
          printf,dun,format='("NAXIS3  ",i5)',nchannel
          free_lun,dun
        endif else begin
          tni = 0l
          tnj = 0l
          tnc = 0l
          fnm = disk+mdl+'/mollosvd'+string(idir,format='(i1)')+'_'+filelabel[im]+'_'+numstr+'.dmp'
          print,'MOLLOSVD: reading file ',fnm
          openr,dun,fnm,/get_lun,/f77
          readu,dun,tni,tnj,tnc
          tintensmapnu = ptr_new(fltarr(tni,tnj,tnc))
          readu,dun,(*tintensmapnu)
          free_lun,dun
          (*intensmapnu)[*,*,*,im] = (*tintensmapnu)[*,*,*]
          ptr_free,tintensmapnu
        endelse
        dv = (maxvph-minvph)/float(nbins)
        for ic = 0,nbins-1 do begin
          histarr[0,ic,im,m] = (0.5+float(ic))*dv + minvph
          for ic1 = 0,nchannel/nbins-1 do begin
            histarr[2,ic,im,m] = histarr[2,ic,im,m] + total((*intensmapnu)[*,*,ic*nchannel/nbins+ic1,im])/float(nar)
          endfor
        endfor
        fwhm[im,m] = sqrt(total(histarr[0,*,im,m]^2*histarr[2,*,im,m])$
                        -(total(histarr[0,*,im,m]*histarr[2,*,im,m])^2/total(histarr[2,*,im,m])))$
                      /sqrt(total(histarr[2,*,im,m]))
        

      endfor
      ptr_free,ttdens,tttemp,ttvlos,tintensnu
    end
    endcase

  endfor


  minv = min(histarr[0,*,*,*])
  maxv = max(histarr[0,*,*,*])
  for im = 0,nmol-1 do begin
    minh[im] = min(histarr[2,*,im,*])
    maxh[im] = max(histarr[2,*,im,*])
  endfor
  dv   = maxv-minv
  minv = minv-0.05*dv
  maxv = maxv+0.05*dv
  dh   = maxh-minh
  minh = minh-0.05*dh
  maxh = maxh+0.05*dh

  dirstr = string(idir+1,format='(i1)')
  set_plot,'ps'
  fnm  = odisk+'mollosvd'+dirstr+'_'+descr+'_vlos.eps'
  print,'MOLLOSVD: writing file ',fnm
  device,filename=fnm,/encaps,/color,bits=8,xsize=15.0,ysize=6.0*nmol
  !p.multi=[0,1,nmol] 
  loadct,ctable
  dypos= 0.9/float(nmol)
  for im = 0,nmol-1 do begin
    if (im eq 0) then txchsz = xchsz else txchsz = 0.0001
    plot,[minv,maxv],[minh[im],maxh[im]],xtitle='!6v!dlos!n [km s!e-1!n]',ytitle='!6T [K] ('+speclabel[im]+'!6)',$
         xcharsize=txchsz,ycharsize=ychsz,$
         xthick=3,ythick=3,charthick=3,/nodata,position=[0.10,0.10+im*dypos,1.00,0.10+(im+1)*dypos],xstyle=1,ystyle=1
    for m=0,models-1 do begin
      oplot,histarr[0,*,im,m],histarr[2,*,im,m],thick=4,psym=10,color=ctype[m]
      legend = "t = "+string(dtdmp*float(dmp[m])*convtime,format='(f4.1)')+' Myr'
      oplot,[minv+dv*0.05,minv+0.15*dv],[maxh[im]-(0.1+0.1*(m+1))*dh[im],maxh[im]-(0.1+0.1*(m+1))*dh[im]],$
            thick=4,color=ctype[m],line=0
      xyouts,minv+0.18*dv,maxh[im]-(0.12+0.1*(m+1))*dh[im],legend,charthick=3,charsize=chsz
      if (finite(fwhm[im,m]) eq 1) then $
        xyouts,minv+0.88*dv,maxh[im]-(0.12+0.1*(m+1))*dh[im],'!6'+string(fwhm[im,m],$
               format='(f4.2)')+' km s!e-1!n',charthick=3,charsize=chsz
    endfor
  endfor
  xyouts,minv+0.05*dv,maxh[nmol-1]-0.1*dh[nmol-1],legend0,charsize=chsz,charthick=3
  !p.multi=0
  device,/close
  set_plot,'x'
 
;===============================================================

  return
end

;===============================================================

function gaussian,nu,temp,mol
; Note that we're using the mol weight of a mol. cloud (2.36), bc
; what counts here is the temperature of the gas
  dnu    = abs(nu-mol.nu0)
  dnud   = (2.80d-7)*mol.nu0*sqrt(temp)
  return,(5.642d-1)*exp(-(dnu/dnud)^2)/dnud 
end

;===============================================================

function getalphanu,nu,dens,temp,mol
  expo   = exp(-(4.798d-11)*mol.nu0/temp)
  anu    = (3.581d19/nu^2) * (mol.abnd*dens/(1d0+mol.ncr/dens+(mol.gu/mol.gl)*expo)) $
           * (mol.gu*mol.a_ul/mol.gl)*(1d0-expo)*gaussian(nu,temp,mol)
  return,anu
end

;===============================================================

pro getintensnu,nx,nchannel,dens,temp,vlos,mol,klo,khi,dx,dnu,minnu,intensnu
  k         = 0l
  nuchan    = 0d0
  nulos     = 0d0
  tnuchan   = 0d0
  tintens   = 0d0
  gtl       = 0d0
  gsf       = 0d0
  alphanu   = ptr_new(dblarr(nx))
  nuaxis    = ptr_new(dblarr(nchannel))
; loop through the spectrum at a given line of sight
  for is=0,nchannel-1 do begin
    tintens= 0d0
    nuchan = (0.5d0 + double(is))*dnu+minnu ; current frequency channel
    (*nuaxis)[is] = nuchan
;   first loop along los: determine alphanu and save it
    for k=klo,khi do begin
      nulos         = mol.nu0*(1d0+(3.33333d-6)*vlos[k])      ; frequency shift due to bulk motion 
      tnuchan       = nuchan + mol.nu0-nulos                  ; shift channel so that it sees the cell at rest
      (*alphanu)[k] = getalphanu(tnuchan,dens[k],temp[k],mol) ; save this. we'll need it twice. This is in cm^{-1}
    endfor
;   second loop along los: determine intensity at current frequency
;   need to walk backwards?
    for k=khi,klo,-1 do begin
;     two steps: get source and tau, do rad trans (see shu), and deal
;     with absorption of cold stuff in front (i.e. the stuff which is
;     important for n<<n_cr)
      gts     = getsource(tnuchan,temp[k])
      gtl     = gettauloc(klo,k,dens,*alphanu,dx,mol)
      tintens = tintens + gts*exp(-gtl)*(*alphanu)[k]*dx
      tintens = tintens * exp(-(*alphanu)[k]*dx)
    endfor
;   the next line assumes Rayleigh-Jeans limit: we're actually
;   returning the brightness temperature (or something like that)
    intensnu[is] = tintens*3.25d36/nuchan^2 
  endfor
  ptr_free,alphanu,nuaxis
  return
end

;=====================================================

function gettauloc,klo,k,dens,alphanu,dx,mol
  return,total(alphanu[klo:k])*dx
end

;=====================================================
 
function getsource,nu,temp
 return,(1.4724d-47)*nu^3/(exp((4.798d-11)*nu/temp)-1d0)
end
